package uk.ac.bbsrc.babraham.SparkSpotter;
/*
 * SparkSpotter - An ImageJ plugin for finding bright sparks in movies
 * Copyright (c) 2007 Simon Andrews

 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Vector;


public class LowessNoramlisation {

	public static int [] normalise (int [] data) {
		// The first thing we need to do is to get an ordered set of measures
		// for the values for this data.

		PixelMeasure [] measures = new PixelMeasure [data.length];
		LowessNoramlisation ssl = new LowessNoramlisation();
		for (int i=0;i<data.length;i++) {
			measures[i] = ssl.new PixelMeasure(data[i],i);
		}
		Arrays.sort(measures);
		
		// We're going to work out the Lowess regression at all points
		
		Vector<double []> lowessPoints = new Vector<double []>();
		
		for (int m=0;m<measures.length;m++) {
			// We need to get the subset of measures we're going to use
			// for this normalisation.  We're using 66% of the data to
			// calculate the smoothing hence all the divide by 3 / 1.5s.
			
			// A nice summary of how to do the lowess curve is at www.itl.nist.gov/div898/handbook/pmd/section1/dep/dep144.htm
			
			
			int startIndex = (int)(m - ((double)measures.length/3));
			int endIndex = (int)(m + ((double)measures.length/3));

			// Correct if we fall off the end
			if (startIndex < 0) {
				startIndex = 0;
//				endIndex = (int)(measures.length/1.5);
			}
			if (endIndex >= measures.length) {
				endIndex = measures.length -1;
//				startIndex = endIndex - (int)(measures.length/2.5);
			}
			
//			System.out.println("For point "+m+" using region "+startIndex+" (a="+measures[startIndex].a+") -"+endIndex+" (a="+measures[endIndex].a+") from "+measures.length);

			// Calculate the correlation coefficient
			
			double maxDist = 0;
			for (int w=startIndex;w<=endIndex;w++) {
				measures[w].distance = measures[w].frame-measures[m].frame;
				if (measures[w].distance < 0) measures[w].distance = 0-measures[w].distance;
				if (measures[w].distance > maxDist) maxDist = measures[w].distance;
			}
//			System.out.println("Calcualted distances");
			
			
			// Scale the distances by the maximum distance in the set and calucalate the weight
			// and do the weighted regression to get the slope and intercept
			double intensityWeightSum=0;
			double frameWeightSum=0;
			double weightSum=0;
			for (int w=startIndex;w<=endIndex;w++) {
				double sd = measures[w].distance/maxDist;
				measures[w].weight= Math.pow((1d - Math.pow(sd,3)),3);
				
				intensityWeightSum += measures[w].intensity * measures[w].weight;
				frameWeightSum += measures[w].frame * measures[w].weight;
				weightSum += measures[w].weight;
			}
			
			double intensityWeightMean = intensityWeightSum/weightSum;
			double frameWeightMean = frameWeightSum/weightSum;
						
			// We can now calculate the slope of the regression.  This is a standard
			// weighted linear regression.
			double slopeNumerator = 0;
			double slopeDenominator = 0;
			
			for (int w=startIndex;w<=endIndex;w++) {
				slopeNumerator += measures[w].weight * ((measures[w].intensity - intensityWeightMean)*(measures[w].frame-frameWeightMean));
				slopeDenominator += measures[w].weight * Math.pow((measures[w].frame - frameWeightMean),2);
			}
			
			double slope = slopeNumerator/slopeDenominator;
			
			
			// And then the intercept
			double intercept = intensityWeightMean - (slope*frameWeightMean);
			
//			System.out.println("Slope="+slope+" intecept="+intercept);
			
			// Finally we can get the corrected control value for this point and assign it
			
			double [] point = new double[2];
			point[0]=(slope*measures[m].frame)+intercept;
			point[1]=measures[m].frame;
//			System.out.println(""+m+"\t"+point[0]+"\t"+point[1]+"\t"+slope);
			lowessPoints.add(point);
			
		}

		// Now we need to go through all of the data correcting the inital
		// values by the corrected values
		
		int [] correctedData = new int [data.length];
		
		Enumeration<double []> e =lowessPoints.elements();
		
		while (e.hasMoreElements()) {
			double [] lp = e.nextElement();
			correctedData[(int)lp[1]] = (int)(data[(int)lp[1]]-lp[0]);
		}
		
		return correctedData;
	
	}
	
	private class PixelMeasure implements Comparable<PixelMeasure> {

		private int intensity;
		private int frame;
		public double weight;
		public int distance;
		
		public PixelMeasure (int intensity, int frame) {
			this.intensity = intensity;
			this.frame = frame;
		}

		public int compareTo(PixelMeasure o) {
			return Double.compare(intensity,o.intensity);
		}
		
	}
	
}
